/*
    Copyright (C) 2008-2019, Millistream Market Data <support@millistream.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef MDF_COMMON_H
# define MDF_COMMON_H 1
# define WINVER 0x0600 /* so we can use some functions on Windows... */
# define _WIN32_WINNT 0x0600 /* so we can use some functions on Windows... */

# ifdef HAVE_CONFIG_H
#  include "../config.h"
# endif
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
# ifdef HAVE_UNISTD_H
#  include <unistd.h>
# endif
# include <stdio.h>
# ifdef HAVE_STDINT_H
#  include <stdint.h>
# elif _MSC_VER >= 1600
#  include <stdint.h>
#  define _STDINT_H 1 /* so we ignore the check in mdf.h */
# elif _MSC_VER
#  define uint8_t unsigned __int8
#  define uint16_t unsigned __int16
#  define uint32_t unsigned __int32
#  define uint64_t unsigned __int64
#  define int64_t __int64
#  define _STDINT_H 1 /* so we ignore the check in mdf.h */
# else
#  define uint8_t unsigned char
#  define uint16_t unsigned short
#  define uint32_t unsigned int
#  define uint64_t unsigned long long
#  define int64_t long long
#  define _STDINT_H 1 /* so we ignore the check in mdf.h */
# endif

# ifdef HAVE_INTTYPES_H
#  include <inttypes.h>
# elif _MSC_VER
#  define PRIu64 "I64u"
# else
#  define PRIu64 "llu"
# endif

# ifndef __UINT64_C
#  ifdef UINT64_C
#   define __UINT64_C(c) UINT64_C(c)
#  elif _MSC_VER
#   define __UINT64_C(c) c##ui64
#  else
#   define __UINT64_C(c) c##ULL
#  endif
# endif

# ifndef HAVE_STRTOULL
#  if _MSC_VER >= 1300
#   define strtoull(a,b,c) _strtoui64(a,b,c)
#  elif HAVE_STRTOQ
#   define strtoull(a,b,c) strtoq(a,b,c)
#  else
#   define strtoull(a,b,c) strtoul(a,b,c)
#  endif
# endif

# ifdef HAVE_STRING_H
#  include <string.h>
# endif
# ifdef HAVE_SYS_IOCTL_H
#  include <sys/ioctl.h>
# endif
# ifdef HAVE_SYS_FILIO_H
#  include <sys/filio.h>
# endif
# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
# ifdef HAVE_SYS_SOCKET_H
#  include <sys/socket.h>
# endif
# ifdef HAVE_SYS_UN_H
#  include <sys/un.h>
# endif
# ifdef HAVE_NETINET_IN_H
#  include <netinet/in.h>
# endif
# ifdef HAVE_NETINET_TCP_H
#  include <netinet/tcp.h>
# endif
# ifdef HAVE_NETDB_H
#  include <netdb.h>
# endif
# ifdef HAVE_FCNTL_H
#  include <fcntl.h>
# endif
# ifdef HAVE_ERRNO_H
#  include <errno.h>
# endif
# ifdef HAVE_SIGNAL_H
#  include <signal.h>
# endif
# ifdef HAVE_TIME_H
#  include <time.h>
# endif
# if defined _WIN32 || defined _WIN64
#  include <stddef.h>
#  include <time.h>
#  include <limits.h>
#  include <winsock2.h>
#  include <ws2tcpip.h>
#  define MSG_NOSIGNAL 0
# else
#  define INVALID_SOCKET -1
# endif
# include <ctype.h>
# include <stdarg.h>
# include <zlib.h>
# include <openssl/opensslv.h>
# include <openssl/rand.h>
# include <openssl/bio.h>
# include <openssl/dh.h>
# include <openssl/rsa.h>
# include <openssl/hmac.h>
# include <openssl/pem.h>
# include <openssl/evp.h>
# include <openssl/engine.h>
# include "mdf_byteswap.h"
# include "mdf.h"
# include "mdf_fields.h"
# include "version.h"
# if _MSC_VER >= 1300
#  define strdup(s) _strdup(s)
# endif
# ifndef SOL_TCP
#  define SOL_TCP IPPROTO_TCP
# endif
# ifndef SOCK_CLOEXEC
#  define SOCK_CLOEXEC 0
# endif

# if (defined LIBRESSL_VERSION_NUMBER && OPENSSL_VERSION_NUMBER == 0x20000000L)
#  undef OPENSSL_VERSION_NUMBER
#  define OPENSSL_VERSION_NUMBER 0x1000107fL
# endif

#define NELEMS(x) ((int)(sizeof x/sizeof x[0]))
#define MAX_RECORDLENGTH (64 * 1024 * 1024)

#define CSTATE_ENCRYPT 1
#define CSTATE_DECRYPT 2
#define CSTATE_INVALID 4 /* prevents mdf_consume() and mdf_message_send() to do anything until the handshake is completed */

struct callbacks {
	void *udata_data;
	void *udata_status;
	mdf_data_callback func_data;
	mdf_status_callback func_status;
};

struct crypto {
	uint8_t *master_secret;
	EVP_CIPHER_CTX *cipher_encrypt_ctx;
	EVP_CIPHER_CTX *cipher_decrypt_ctx;
	EVP_MD_CTX *digest_encrypt_ctx;
	EVP_MD_CTX *digest_decrypt_ctx;
	const EVP_MD *digest;
	uint64_t encrypt_cnt;
	uint64_t decrypt_cnt;
	int cstate;
	int digest_length;
};

struct mdf_s {
	uint64_t bytes_read;
	uint64_t bytes_written;
	uint64_t template_version;
	uint8_t *data;
	uint8_t *value;
	struct crypto crypto;
	struct msg_template *templates;
	struct msg_template *template;
	struct field **sendfields;
	struct field **sendtags;
	uint8_t *sendbuf;
	char *connected_host;
	char *connected_ip;
	char *bind_address;
	struct callbacks callbacks;
	time_t time_lastupdate;
	size_t pos; /* points at the current pmap */
	size_t fpos; /* points at the current field */
	size_t vlen; /* length of value in bytes */
	size_t data_pos; /* current decoding position in the data buffer */
	size_t data_used; /* current read position in the data buffer */
	size_t data_size; /* total size of the data buffer */
	size_t message_len; /* the total message length including any message digest */
	size_t message_end; /* points to the last byte of the message (not including the digest) */
	int templates_num;
	int sendfields_num;
	unsigned int sendtags_num;
	int sendbuf_size;
	int in_pmap_pos; /* bit position in the current pmap, 128,96,64,32,16,8,4,1 */
	int field_pmap_pos; /* current field position */
	int fstatus;
	int connect_timeout;
	int heartbeats_sent;
	int heartbeat_interval;
	int max_missed_heartbeats;
	int tcp_nodelay;
	int no_encryption;
	int timediff;
	MDF_ERROR error;
# if defined _WIN32 || defined _WIN64
	SOCKET fd;
# else
	int fd;
# endif
	uint8_t oldval; /* carries the byte that we zero out when reading messages */
};

void mdf_int_send_rekey (mdf_t handle);
int mdf_int_send_heartbeat (mdf_t handle);

#endif
